# coding: utf-8


class Transformer(object):
    def prevalidate_request(self, slicer, metrics, dimensions,
                            metric_filters, dimension_filters,
                            references, operations):
        pass

    def transform(self, dataframe, display_schema):
        raise NotImplementedError


class TransformationException(Exception):
    pass



# coding: utf-8
from fireant.slicer import transformers
from fireant.slicer.managers import SlicerManager, TransformerManager
from pypika import JoinType
from pypika.terms import Mod


class SlicerElement(object):
    """The `SlicerElement` class represents an element of the slicer, either a metric or dimension, which contains
    information about such as how to query it from the database."""

    def __init__(self, key, label=None, definition=None, joins=None):
        """
        :param key:
            The unique identifier of the slicer element, used in the Slicer manager API to reference a defined element.

        :param label:
            A displayable representation of the column.  Defaults to the key capitalized.

        :param definition:
            The definition of the element as a PyPika expression which defines how to query it from the database.

        :param joins:
            A list of Join keys required by this element.
        """
        self.key = key
        self.label = label or key
        self.definition = definition
        self.joins = joins

    def __unicode__(self):
        return self.key

    def __repr__(self):
        return self.key

    def schemas(self, *args, **kwargs):
        return [
            (self.key, self.definition)
        ]


class Metric(SlicerElement):
    """
    The `Metric` class represents a metric in the `Slicer` object.
    """

    def __init__(self, key, label=None, definition=None, joins=None, precision=None, prefix=None, suffix=None):
        super(Metric, self).__init__(key, label, definition, joins)
        self.precision = precision
        self.prefix = prefix
        self.suffix = suffix


class Dimension(SlicerElement):
    """
    The `Dimension` class represents a dimension in the `Slicer` object.
    """

    def __init__(self, key, label=None, definition=None, joins=None):
        super(Dimension, self).__init__(key, label, definition, joins)

    def levels(self):
        return [self.key]


class NumericInterval(object):
    def __init__(self, size=1, offset=0):
        self.size = size
        self.offset = offset

    def __eq__(self, other):
        return isinstance(other, NumericInterval) and self.size == other.size and self.offset == other.offset

    def __hash__(self):
        return hash('NumericInterval %d %d' % (self.size, self.offset))

    def __str__(self):
        return 'NumericInterval(size=%d,offset=%d)' % (self.size, self.offset)


class ContinuousDimension(Dimension):
    def __init__(self, key, label=None, definition=None, default_interval=NumericInterval(1, 0), joins=None):
        super(ContinuousDimension, self).__init__(key=key, label=label, definition=definition, joins=joins)
        self.default_interval = default_interval

    def schemas(self, *args, **kwargs):
        size, offset = args if args else (self.default_interval.size, self.default_interval.offset)
        return [(self.key, Mod(self.definition + offset, size))]


class DatetimeInterval(object):
    def __init__(self, size):
        self.size = size

    def __eq__(self, other):
        return isinstance(other, DatetimeInterval) and self.size == other.size

    def __hash__(self):
        return hash('DatetimeInterval %s' % self.size)

    def __str__(self):
        return 'DatetimeInterval(interval=%s)' % self.size


class DatetimeDimension(ContinuousDimension):
    hour = DatetimeInterval('HH')
    day = DatetimeInterval('DD')
    week = DatetimeInterval('IW')
    month = DatetimeInterval('MM')
    quarter = DatetimeInterval('Q')
    year = DatetimeInterval('Y')

    def __init__(self, key, label=None, definition=None, default_interval=day, joins=None):
        super(DatetimeDimension, self).__init__(key=key, label=label, definition=definition, joins=joins,
                                                default_interval=default_interval)

    def schemas(self, *args, **kwargs):
        interval = args[0] if args else self.default_interval
        return [(self.key, kwargs['database'].trunc_date(self.definition, interval.size))]


class CategoricalDimension(Dimension):
    def __init__(self, key, label=None, definition=None, display_options=tuple(), joins=None):
        super(CategoricalDimension, self).__init__(key=key, label=label, definition=definition, joins=joins)
        self.display_options = display_options


class UniqueDimension(Dimension):
    def __init__(self, key, label=None, definition=None, display_field=None, joins=None):
        super(UniqueDimension, self).__init__(key=key, label=label, definition=definition, joins=joins)
        self.display_field = display_field

    def display_key(self):
        return '{key}_display'.format(key=self.key)

    def schemas(self, *args, **kwargs):
        schemas = [('{key}'.format(key=self.key), self.definition)]

        if self.display_field:
            schemas.append((self.display_key(), self.display_field))

        return schemas

    def levels(self):
        if self.display_field is not None:
            return [self.key, self.display_key()]
        return super(UniqueDimension, self).levels()


class BooleanDimension(Dimension):
    def __init__(self, key, label=None, definition=None, joins=None):
        super(BooleanDimension, self).__init__(key=key, label=label, definition=definition, joins=joins)


class DimensionValue(object):
    """
    An option belongs to a categorical dimension which specifies a fixed set of values
    """

    def __init__(self, key, label=None):
        self.key = key
        self.label = label or key


class Join(object):
    def __init__(self, key, table, criterion, join_type=JoinType.inner):
        self.key = key
        self.table = table
        self.criterion = criterion
        self.join_type = join_type


class Slicer(object):
    def __init__(self, table, database, metrics=tuple(), dimensions=tuple(), joins=tuple(), hint_table=None):
        """
        Constructor for a slicer.  Contains all the fields to initialize the slicer.

        :param table: (Required)
            A Pypika Table reference. The primary table that this slicer will retrieve data from.

        :param database:  (Required)
            A Database reference. Holds the connection details used by this slicer to execute queries.

        :param metrics: (Required: At least one)
            A list of metrics which can be queried.  Metrics are the types of data that are displayed.

        :param dimensions: (Optional)
            A list of dimensions used for grouping metrics.  Dimensions are used as the axes in charts, the indices in
            tables, and also for splitting charts into multiple lines.

        :param joins:  (Optional)
            A list of join descriptions for joining additional tables.  Joined tables are only used when querying a
            metric or dimension which requires it.

        :param hint_table: (Optional)
            A hint table used for querying dimension options.  If not present, the table will be used.  The hint_table
            must have the same definition as the table omitting dimensions which do not have a set of options (such as
            datetime dimensions) and the metrics.  This is provided to more efficiently query dimension options.
        """
        self.table = table
        self.database = database

        self.metrics = {metric.key: metric for metric in metrics}
        self.dimensions = {dimension.key: dimension for dimension in dimensions}
        self.joins = {join.key: join for join in joins}
        self.hint_table = hint_table

        self.manager = SlicerManager(self)
        for name, bundle in transformers.BUNDLES.items():
            setattr(self, name, TransformerManager(self.manager, bundle))


class EqualityOperator(object):
    eq = 'eq'
    ne = 'ne'
    gt = 'gt'
    lt = 'lt'
    gte = 'gte'
    lte = 'lte'



# coding: utf8
from fireant import utils


class Filter(object):
    def __init__(self, element_key):
        self.element_key = element_key


class EqualityFilter(Filter):
    def __init__(self, element_key, operator, value):
        super(EqualityFilter, self).__init__(element_key)
        self.operator = operator
        self.value = value

    def schemas(self, elements, **kwargs):
        elements = utils.wrap_list(elements)
        value = utils.wrap_list(self.value)

        criterion = None
        for element, value in zip(elements, value):
            crit = getattr(element, self.operator)(value)
            if criterion:
                criterion = criterion & crit
            else:
                criterion = crit

        return criterion

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and self.element_key == other.element_key \
               and self.operator == other.operator \
               and self.value == other.value


class ContainsFilter(Filter):
    def __init__(self, element_key, values):
        super(ContainsFilter, self).__init__(element_key)
        self.values = values

    def schemas(self, element):
        return element.isin(self.values)

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and self.element_key == other.element_key \
               and self.values == other.values


class ExcludesFilter(Filter):
    def __init__(self, element_key, values):
        super(ExcludesFilter, self).__init__(element_key)
        self.values = values

    def schemas(self, element):
        return element.notin(self.values)

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and self.element_key == other.element_key \
               and self.values == other.values


class RangeFilter(Filter):
    def __init__(self, element_key, start, stop):
        super(RangeFilter, self).__init__(element_key)
        self.start = start
        self.stop = stop

    def schemas(self, element):
        element = utils.wrap_list(element)
        starts = utils.wrap_list(self.start)
        stops = utils.wrap_list(self.stop)

        criterion = None
        for el, start, stop in zip(element, starts, stops):
            crit = el[start:stop]
            if criterion:
                criterion = criterion & crit
            else:
                criterion = crit

        return criterion

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and self.element_key == other.element_key \
               and self.start == other.start \
               and self.stop == other.stop


class WildcardFilter(Filter):
    def __init__(self, element_key, value):
        super(WildcardFilter, self).__init__(element_key)
        self.value = value

    def schemas(self, element):
        return element.like(self.value)

    def __eq__(self, other):
        return isinstance(other, self.__class__) \
               and self.element_key == other.element_key \
               and self.value == other.value



# coding: utf-8


def dimension_levels(dimension_key, dimension):
    if 'display_field' in dimension:
        return [dimension_key, dimension['display_field']]
    return [dimension_key]


def wrap_list(value):
    return value if isinstance(value, (tuple, list)) else [value]


def flatten(items):
    return [item for level in items for item in wrap_list(level)]


def slice_first(item):
    if isinstance(item, (tuple, list)):
        return item[0]
    return item


def filter_duplicates(iterable):
    filtered_list, seen = [], set()
    for item in iterable:
        key = slice_first(item)

        if key in seen:
            continue

        seen.add(key)
        filtered_list.append(item)

    return filtered_list



# coding: utf-8
from pypika import terms

from . import Database


class Trunc(terms.Function):
    # Wrapper for Vertica TRUNC function for truncating dates.

    def __init__(self, field, date_format, alias=None):
        super(Trunc, self).__init__('TRUNC', field, date_format, alias=alias)


class Interval(terms.Interval):
    # Wrapper for Vertica INTERVAL term.

    def get_sql(self):
        if hasattr(self, 'quarters'):
            expr = getattr(self, 'quarters')
            unit = 'QUARTER'

        elif hasattr(self, 'weeks'):
            expr = getattr(self, 'weeks')
            unit = 'WEEK'

        else:
            # Create the whole expression but trim out the unnecessery fields
            expr = self.trim_pattern.sub(
                '',
                "{years}-{months}-{days} {hours}:{minutes}:{seconds}.{microseconds}".format(
                    years=getattr(self, 'years', 0),
                    months=getattr(self, 'months', 0),
                    days=getattr(self, 'days', 0),
                    hours=getattr(self, 'hours', 0),
                    minutes=getattr(self, 'minutes', 0),
                    seconds=getattr(self, 'seconds', 0),
                    microseconds=getattr(self, 'microseconds', 0),
                )
            )
            unit = '{largest}_{smallest}'.format(
                largest=self.largest,
                smallest=self.smallest,
            ) if self.largest != self.smallest else self.largest

        if unit in ['YEAR']:
            query = 'INTERVAL \'{expr}\' {unit}'
        else:
            # This has day as its maximum interval granularity, so it does not take into account leap years.
            query = 'INTERVAL \'{expr} {unit}\''

        return query.format(expr=expr,  unit=unit)



class Vertica(Database):
    # Vertica client that uses the vertica_python driver.

    def __init__(self, host='localhost', port=5433, database='vertica',
                 user='vertica', password=None,
                 read_timeout=None):
        self.host = host
        self.port = port
        self.database = database
        self.user = user
        self.password = password
        self.read_timeout = read_timeout

    def connect(self):
        import vertica_python

        return vertica_python.connect(
            host=self.host, port=self.port, database=self.database,
            user=self.user, password=self.password,
            read_timeout=self.read_timeout,
        )

    def trunc_date(self, field, interval):
        return Trunc(field, interval)

    def interval(self, **kwargs):
        return Interval(**kwargs)



# coding: utf-8
import pandas as pd

from fireant import settings
from . import Transformer, TransformationException


def _format_dimension_labels(dimension):
    if 'display_field' in dimension:
        return ['%s ID' % dimension['label'], dimension['label']]

    return [dimension['label']]


class PandasRowIndexTransformer(Transformer):
    def transform(self, dataframe, display_schema):
        dataframe = self._set_display_options(dataframe, display_schema)
        if display_schema['dimensions']:
            dataframe = self._set_dimension_labels(dataframe, display_schema['dimensions'])
        dataframe = self._set_metric_labels(dataframe, display_schema['metrics'], display_schema.get('references'))

        if isinstance(dataframe.index, pd.MultiIndex):
            drop_levels = ['%s ID' % dimension['label']
                           for key, dimension in display_schema['dimensions'].items()
                           if dimension.get('display_field')]
            dataframe.reset_index(level=drop_levels, drop=True, inplace=True)

        return dataframe

    def _set_display_options(self, dataframe, display_schema):
        dataframe = dataframe.copy()

        for key, dimension in display_schema['dimensions'].items():
            if 'display_options' in dimension:
                display_values = [dimension['display_options'].get(value, value)
                                  for value in dataframe.index.get_level_values(key).unique()]

                if isinstance(dataframe.index, pd.MultiIndex):
                    dataframe.index.set_levels(display_values, key, inplace=True)

                else:
                    dataframe.index = pd.Index(display_values)

        return dataframe

    def _set_dimension_labels(self, dataframe, dimensions):
        dataframe = dataframe.copy()
        dataframe.index.names = [label
                                 for key, dimension in dimensions.items()
                                 for label in _format_dimension_labels(dimension)]

        return dataframe

    def _set_metric_labels(self, dataframe, metrics, references):
        dataframe = dataframe.copy()

        labels = [metric['label'] for metric in metrics.values()]
        if isinstance(dataframe.columns, pd.MultiIndex):
            dataframe.columns = dataframe.columns.reorder_levels((1, 0)) \
                .set_levels([labels,
                             [None] + list(references.values())]) \
                .set_names('Reference', 1)

        else:
            dataframe.columns = labels

        return dataframe

class PandasColumnIndexTransformer(PandasRowIndexTransformer):
    def transform(self, dataframe, display_schema):
        dataframe = super(PandasColumnIndexTransformer, self).transform(dataframe, display_schema)

        if isinstance(dataframe.index, pd.MultiIndex):
            drop_levels = [dimension['label']
                           for dimension in list(display_schema['dimensions'].values())[1:]]

            dataframe = dataframe.unstack(level=drop_levels)

        return dataframe


class MatplotlibLineChartTransformer(PandasColumnIndexTransformer):
    def transform(self, dataframe, display_schema):
        self._validate_dimensions(dataframe, display_schema['dimensions'])
        dataframe = super(MatplotlibLineChartTransformer, self).transform(dataframe, display_schema)

        metrics = list(display_schema['metrics'].values())
        height, width = settings.matplotlib_figsize or (14, 5)
        figsize = (height, width * len(metrics))

        if 1 == len(metrics):
            return dataframe.plot.line(figsize=figsize) \
                .legend(loc='center left', bbox_to_anchor=(1, 0.5)) \
                .set_title(metrics[0]['label'])

        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise TransformationException('Missing library: matplotlib')

        fig, axes = plt.subplots(len(metrics), sharex=True, figsize=figsize)
        for metric, axis in zip(metrics, axes):
            label = metric['label']
            dataframe[label].plot.line(ax=axis) \
                .legend(loc='center left', bbox_to_anchor=(1, 0.5)) \
                .set_title(label)

        return axes

    def _validate_dimensions(self, dataframe, dimensions):
        if not 0 < len(dimensions):
            raise TransformationException('Cannot transform line chart.  '
                                          'At least one dimension is required.')


class MatplotlibBarChartTransformer(PandasColumnIndexTransformer):
    def transform(self, dataframe, display_schema):
        dataframe = super(MatplotlibBarChartTransformer, self).transform(dataframe, display_schema)

        return dataframe.plot.bar(figsize=settings.matplotlib_figsize or (14, 5)) \
            .legend(loc='center left', bbox_to_anchor=(1, 0.5))



# coding: utf-8
import copy
import logging
import time
from itertools import chain

import pandas as pd

from fireant import utils
from fireant.slicer.references import YoY, Delta, DeltaPercentage
from pypika import Query, JoinType

query_logger = logging.getLogger('fireant.query_log$')


class QueryManager(object):
    def query_data(self, database, table, joins=None,
                   metrics=None, dimensions=None,
                   mfilters=None, dfilters=None,
                   references=None, rollup=None):
        """
        Loads a pandas data frame given a table and a description of the request.

        :param database:
            The database interface to use to execute the connection

       :param table:
            Type: pypika.Table
            (Required) The primary table to select data from.  In SQL, this is the table in the FROM clause.

        :param joins:
            Type: list[tuple[pypika.Table, pypika.Criterion, pypika.JoinType]]
            (Optional) A list of tables to join in the query.  If a metric should be selected from a table besides the
            table parameter, the additional table *must* be joined.

            The tuple includes in the following order: Join table, Join criterion, Join type

        :param metrics:
            Type: dict[str: pypika.Field]]
            A dict containing metrics fields to query.  This value is required to contain at least one entry.  The key
            used will match the corresponding column in the returned pd.DataFrame.

        :param dimensions:
            Type: OrderedDict[str: pypika.Field]]
            (Optional) An OrderedDict containing indices to query.  If empty, a pd.Series will be returned containing
            one value per metric.  If given one entry, a pd.DataFrame with a single index will be returned.  If more
            than one value is given, then a pd.DataFrame with a multi-index will be returned.  The key used will match
            the corresponding index level in the returned DataFrame.

            If a dict is used instead of an OrderedDict, then the index levels cannot be guarenteed, so in most cases an
            OrderedDict should be used.

        :param mfilters:
            Type: list[pypika.Criterion]
            (Optional) A list containing criterions to use in the HAVING clause of the query.  This will reduce the data
            to instances where metrics meet the criteria.

        :param dfilters:
            Type: list[pypika.Criterion]
            (Optional) A list containing criterions to use in the WHERE clause of the query.  This will reduce the data
            to instances where dimensions meet the criteria.

        :param references:
            Type: dict[str: str]]
            (Optional) A dict containing comparison operations.  The key must match a valid reference.  The value must
            match the key of a provided dimension.  The value must match a dimensions of the appropriate type, usually a
            Time Series dimension.

            For each value passed, the query will be duplicated containing values for a comparison.

            References can also modify the comparison field to show values such as change.  The format is as follows:

            {key}_{modifier}

            The modifier type is optional and the underscore must be omitted when one is not used.

            Valid references are as follows:

            - *wow*: Week over Week - duplicates the query for values one week prior.
            - *mom*: Month over Month - duplicates the query for values 4 weeks prior.
            - *qoq*: Quarter over Quarter - duplicates the query for values one quarter prior.
            - *yoy*: Year over Year - duplicates the query for values 52 weeks prior.

            Valid modifier types are as follows:

            - *d*: Delta - difference between previous value and current
            - *p*: Delta Percentage - difference between previous value and current as a percentage of the previous
                    value.

            Examples:

            "wow":   Week over Week
            "wow_d": delta Week over Week

        :param rollup:
            Type: list[str]
            (Optional) Lists the dimensions to include for ROLLUP which calculates the totals across groups.  This list
            must be a subset of the keys of the dimensions parameter dict.

            .. note::

                When using rollup for less than all of the dimensions, the dimensions included in the ROLLUP will be
                moved after the non-ROLLUP dimensions.

        :return:
            A pd.DataFrame indexed by the provided dimensions paramaters containing columns for each metrics parameter.
        """
        query = self._build_data_query(
            database, table, joins, metrics, dimensions, dfilters, mfilters, references, rollup
        )
        query_string = str(query)

        start_time = time.time()
        dataframe = database.fetch_dataframe(query_string)

        query_logger.info('[duration: {duration} seconds]: {query}'.format(
            duration=round(time.time() - start_time, 4),
            query=query_string)
        )

        dataframe.columns = [col.decode('utf-8') if isinstance(col, bytes) else col
                             for col in dataframe.columns]

        for column in dataframe.columns:
            # Only replace NaNs for columns of type object. Column types other than that tend to be checked
            # against in the transformers. Which would be a problem when replacing NaNs with a string
            # because that alters the type of the column.
            if dataframe[column].dtype == object:
                dataframe[column] = dataframe[column].fillna('')

        if dimensions:
            dataframe = dataframe.set_index(
                # Removed the reference keys for now
                list(dimensions.keys())  # + ['{1}_{0}'.format(*ref) for ref in references.items()]
            ).sort_index()

        if references:
            dataframe.columns = pd.MultiIndex.from_product([[''] + list(references.keys()), list(metrics.keys())])

        return dataframe.fillna(0)

    def query_dimension_options(self, database, table, joins=None, dimensions=None, filters=None, limit=None):
        """
        Builds and executes a query to retrieve possible dimension options given a set of filters.

        :param database:
            The database interface to use to execute the connection
        :param table:
            See above

        :param joins:
            See above

        :param dimensions:
            See above

        :param filters:
            See above

        :param limit:
            An optional limit to the number of results returned.

        :return:
        """
        query = self._build_dimension_query(table, joins, dimensions, filters, limit)
        results = database.fetch(str(query))
        return [{k: v for k, v in zip(dimensions.keys(), result)}
                for result in results]

    def _build_data_query(self, database, table, joins, metrics, dimensions, dfilters, mfilters, references, rollup):
        args = (table, joins or dict(), metrics or dict(), dimensions or dict(),
                dfilters or dict(), mfilters or dict(), rollup or list())

        query = self._build_query_inner(*args)

        if references:
            return self._build_reference_query(query, database, references, *args)

        return self._add_sorting(query, list(dimensions.values()))

    def _build_reference_query(self, query, database, references, table, joins, metrics, dimensions, dfilters,
                               mfilters, rollup):
        wrapper_query = Query.from_(query).select(*[query.field(key).as_(key)
                                                    for key in list(dimensions.keys()) + list(metrics.keys())])

        for reference_key, schema in references.items():
            dimension_key, interval = schema['dimension'], schema['interval']

            # The interval term from pypika does not take into account leap years, therefore the interval
            # needs to be replaced with a database specific one when appropriate.
            if reference_key in [YoY.key, Delta.generate_key(YoY.key), DeltaPercentage.generate_key(YoY.key)]:
                interval = database.interval(years=1)

            schema['interval'] = interval

            # Don't reuse the dfilters arg otherwise intervals will be aggregated on each iteration
            replaced_dfilters = self._replace_filters_for_ref(dfilters, schema['definition'], interval)
            ref_query = self._build_query_inner(table, joins, metrics, dimensions, replaced_dfilters, mfilters, rollup)
            join_criteria = self._build_reference_join_criteria(dimension_key, dimensions, interval, query, ref_query)

            # Optional modifier function to modify the metric in the reference query. This is for delta and delta
            # percentage references. It is None for normal references and this default should be used
            modifier = schema.get('modifier')

            # Join the reference query and select the reference dimension and all metrics
            # This ignores additional dimensions since they are identical to the primary results
            if join_criteria:
                wrapper_query = wrapper_query.join(ref_query, JoinType.left).on(join_criteria)
            else:
                wrapper_query = wrapper_query.from_(ref_query)

            if modifier:
                get_reference_field = lambda key: modifier(query.field(key), ref_query.field(key))

            else:
                get_reference_field = lambda key: ref_query.field(key)

            wrapper_query = wrapper_query.select(
                *[get_reference_field(key).as_(self._suffix(key, reference_key))
                  for key in metrics.keys()]
            )

        return self._add_sorting(wrapper_query, [query.field(dkey) for dkey in dimensions.keys()])

    def _build_dimension_query(self, table, joins, dimensions, filters, limit=None):
        query = Query.from_(table).distinct()
        query = self._add_joins(joins, query)
        query = self._add_filters(query, filters, [])

        for key, dimension in dimensions.items():
            query = query.select(dimension.as_(key))

        if limit:
            query = query[:limit]

        return query

    def _build_query_inner(self, table, joins, metrics, dimensions, dfilters, mfilters, rollup):
        query = Query.from_(table)
        query = self._add_joins(joins, query)
        query = self._select_dimensions(query, dimensions, rollup)
        query = self._select_metrics(query, metrics)
        return self._add_filters(query, dfilters, mfilters)

    @staticmethod
    def _add_joins(joins, query):
        for join_table, criterion, join_type in joins:
            query = query.join(join_table, how=join_type).on(criterion)
        return query

    @staticmethod
    def _select_dimensions(query, dimensions, rollup):
        dims = [dimension.as_(key)
                for key, dimension in dimensions.items()
                if key not in chain(*rollup)]
        if dims:
            query = query.select(*dims).groupby(*dims)

        # Rollup is passed in as a list of lists so that multiple columns can be rolled up together (such as for
        # Unique dimensions)
        rollup_dims = [[dimension.as_(dimension_key)
                        for dimension_key, dimension in dimensions.items()
                        if dimension_key in keys]
                       for keys in rollup]

        # Remove entry levels
        flattened_rollup_dims = utils.flatten(rollup_dims)

        if flattened_rollup_dims:
            query = query.select(*flattened_rollup_dims).rollup(*rollup_dims)

        return query

    @staticmethod
    def _select_metrics(query, metrics):
        return query.select(*[metric.as_(key)
                              for key, metric in metrics.items()])

    @staticmethod
    def _add_filters(query, dfilters, mfilters):
        if dfilters:
            for dfx in dfilters:
                query = query.where(dfx)

        if mfilters:
            for mfx in mfilters:
                query = query.having(mfx)

        return query

    @staticmethod
    def _add_sorting(query, dimensions):
        if dimensions:
            return query.orderby(*dimensions)
        return query

    @staticmethod
    def _suffix(key, suffix):
        return '%s_%s' % (key, suffix) if suffix else key

    @staticmethod
    def _replace_filters_for_ref(dfilters, target_dimension, interval):
        """
        Replaces the dimension used by a reference in the dimension schema and dimension filter schema.

        We do this in order to build the same query with a shifted date instead of the actual date.
        """
        new_dfilters = []
        for dfilter in dfilters:
            try:
                # FIXME this is a bit hacky. Casts the fields to string to see if the filter uses this dimensinon
                # TODO provide a utility in pypika for checking if these are the same
                filter_fields = ''.join(str(f) for f in dfilter.term.fields())
                target_fields = ''.join(str(f) for f in target_dimension.fields())
                if filter_fields == target_fields:
                    dfilter = copy.deepcopy(dfilter)
                    dfilter.term = dfilter.term + interval

            except (AttributeError, IndexError):
                pass  # If the above if-expression cannot be evaluated, then its not the filter we are looking for

            new_dfilters.append(dfilter)

        return new_dfilters

    @staticmethod
    def _build_reference_join_criteria(dimension_key, dimensions, interval, query, ref_query):
        ref_join_criteria = []
        if dimension_key in dimensions:
            ref_join_criteria.append(
                query.field(dimension_key) == ref_query.field(dimension_key) + interval
            )
        ref_join_criteria += [query.field(dkey) == ref_query.field(dkey)
                              for dkey in set(dimensions.keys()) - {dimension_key}]

        # If there are no selected dimensions, this will be an empty list.
        if not ref_join_criteria:
            return None

        ref_criteria = ref_join_criteria.pop(0)
        for criteria in ref_join_criteria:
            ref_criteria &= criteria

        return ref_criteria



# coding: utf-8
import locale as lc
from datetime import time

import numpy as np
import pandas as pd
from fireant import settings
from fireant.slicer.operations import Totals
from fireant.slicer.transformers import Transformer

NO_TIME = time(0)


def _safe(value):
    if isinstance(value, pd.Timestamp):
        if value.time() == NO_TIME:
            return value.strftime('%Y-%m-%d')
        else:
            return value.strftime('%Y-%m-%dT%H:%M:%S')

    if value is None or (isinstance(value, float) and np.isnan(value)) or pd.isnull(value):
        return None

    if isinstance(value, np.int64):
        # Cannot transform np.int64 to json
        return int(value)

    if isinstance(value, np.float64):
        return float(value)

    return value


def _pretty(value, schema):
    precision = schema.get('precision')

    if isinstance(value, float):
        if precision is not None:
            float_format = '%d' if precision == 0 else '%.{}f'.format(precision)
            value = lc.format(float_format, value, grouping=True)
        elif value.is_integer():
            float_format = '%d'
            value = lc.format(float_format, value, grouping=True)
        else:
            float_format = '%f'
            # Stripping trailing zeros is necessary because %f can add them if no precision is set
            value = lc.format(float_format, value, grouping=True).rstrip('.0')

    if isinstance(value, int):
        float_format = '%d'
        value = lc.format(float_format, value, grouping=True)

    return '{prefix}{value}{suffix}'.format(
        prefix=schema.get('prefix', ''),
        value=str(value),
        suffix=schema.get('suffix', ''),
    )


def _format_value(value, metric):
    raw_value = _safe(value)
    return {'value': raw_value, 'display': _pretty(raw_value, metric) if raw_value is not None else None}


def _format_column_display(csv_df, metrics, dimensions):
    column_display = []
    for idx in list(csv_df.columns):
        metric_value, dimension_values = idx[0], idx[1:]

        dimension_displays = []
        for dimension_level, dimension_value in zip(csv_df.columns.names[1:], dimension_values):
            if 'display_options' in dimensions[dimension_level]:
                dimension_display = dimensions[dimension_level]['display_options'].get(dimension_value,
                                                                                       dimension_value)
            else:
                dimension_display = dimension_value

            dimension_display = _safe(dimension_display)

            if dimension_display != Totals.label:
                dimension_displays.append(dimension_display)

        if dimension_displays:
            column_display.append('{metric} ({dimensions})'.format(
                metric=metrics[metric_value]['label'],
                dimensions=', '.join(dimension_displays),
            ))
        else:
            column_display.append(metrics[metric_value]['label'])

    return column_display


class DataTablesRowIndexTransformer(Transformer):
    def transform(self, dataframe, display_schema):
        dataframe = self._prepare_dataframe(dataframe, display_schema['dimensions'])

        return {
            'columns': self._render_columns(dataframe, display_schema),
            'data': self._render_data(dataframe, display_schema),
        }

    def _prepare_dataframe(self, dataframe, dimensions):
        # Replaces invalid values and unstacks the data frame for column_index tables.
        return dataframe.replace([np.inf, -np.inf], np.nan)

    def _render_columns(self, dataframe, display_schema):
        maxcols = settings.datatables_maxcols
        dimension_columns = [self._render_dimension_column(dimension_key, display_schema)
                             for dimension_key in dataframe.index.names[:maxcols]
                             if dimension_key in display_schema['dimensions']]

        maxcols -= len(dimension_columns)
        metric_columns = [self._render_column_level(metric_column, display_schema)
                          for metric_column in list(dataframe.columns)[:maxcols]]

        return dimension_columns + metric_columns

    def _render_dimension_column(self, key, display_schema):
        dimension = display_schema['dimensions'][key]

        if 'display_field' in dimension or 'display_options' in dimension:
            render = {
                '_': 'display',
                'type': 'display',
                'sort': 'display'
            }
        else:
            render = {
                '_': 'value',
                'type': 'value',
                'sort': 'value'
            }

        return {'title': dimension['label'],
                'data': '{}'.format(key),
                'render': render}

    def _render_column_level(self, metric_column, display_schema):
        metrics = display_schema['metrics']
        if not isinstance(metric_column, tuple):
            return {'title': metrics[metric_column]['label'],
                    'data': metric_column,
                    'render': {'type': 'value', '_': 'display', 'sort': 'value'}}

        references = display_schema.get('references')
        metric_key_idx = 1 if references else 0
        if metric_key_idx and metric_column[0]:
            reference_key = metric_column[0]
            metric_key = metric_column[1]
            data_keys = [reference_key, metric_key]
            metric_label = '{metric} {reference}'.format(
                metric=metrics[metric_key]['label'],
                reference=display_schema['references'][reference_key]
            )

        else:
            metric_key = metric_column[metric_key_idx]
            data_keys = [metric_key]
            metric_label = metrics[metric_key]['label']

        path = '.'.join(map(str, data_keys))
        return {
            'title': metric_label,
            'data': path,
            'render': {'type': 'value', '_': 'display', 'sort': 'value'}
        }

    def _render_data(self, dataframe, display_schema):
        n = len(dataframe.index.levels) if isinstance(dataframe.index, pd.MultiIndex) else 1
        dimensions = list(display_schema['dimensions'].items())
        row_dimensions, column_dimensions = dimensions[:n], dimensions[n:]

        data = []
        for idx, df_row in dataframe.iterrows():
            row = {}

            if not isinstance(idx, tuple):
                idx = (idx,)

            for key, value in self._render_dimension_data(idx, row_dimensions):
                row[key] = value

            for key, value in self._render_metric_data(df_row, column_dimensions,
                                                       display_schema['metrics'], display_schema.get('references')):
                row[key] = value

            data.append(row)

        return data

    def _render_dimension_data(self, idx, dimensions):
        i = 0
        for key, dimension in dimensions:
            dimension_value = _safe(idx[i])

            if 'display_field' in dimension:
                i += 1
                yield key, {'display': _safe(idx[i]), 'value': dimension_value}

            elif 'display_options' in dimension:
                display = dimension['display_options'].get(dimension_value, dimension_value)
                yield key, {'display': display, 'value': dimension_value}

            else:
                yield key, {'value': dimension_value}

            i += 1

    def _render_metric_data(self, dataframe, dimensions, metrics, references):
        if not isinstance(dataframe.index, pd.MultiIndex):
            for metric_key, label in metrics.items():
                yield metric_key, _format_value(dataframe[metric_key], metrics[metric_key])

        if references:
            for reference in [''] + list(references):
                for key, value in self._recurse_dimensions(dataframe[reference], dimensions, metrics,
                                                           reference or None):
                    yield key, value
            return

        for key, value in self._recurse_dimensions(dataframe, dimensions, metrics):
            yield key, value

    def _recurse_dimensions(self, dataframe, dimensions, metrics, reference=None):
        if reference:
            return [(reference, dict(self._recurse_dimensions(dataframe, dimensions, metrics)))]

        return [(metric_key, _format_value(dataframe[metric_key], metrics[metric_key]))
                for metric_key, label in metrics.items()]


class DataTablesColumnIndexTransformer(DataTablesRowIndexTransformer):
    def _prepare_dataframe(self, dataframe, dimensions):
        # Replaces invalid values and unstacks the data frame for column_index tables.
        dataframe = super(DataTablesColumnIndexTransformer, self)._prepare_dataframe(dataframe, dimensions)

        unstack_levels = []
        for key, dimension in list(dimensions.items())[1:]:
            unstack_levels.append(key)
            if 'display_field' in dimension:
                unstack_levels.append(dimension['display_field'])

        dataframe.replace('', 'N/A', inplace=True)

        # TODO: Remove the extra NaN row that is created for each column when totals are being applied to a
        #       unique dimension.
        return dataframe.unstack(level=unstack_levels, fill_value='N/A')

    def _render_column_level(self, metric_column, display_schema):
        column = super(DataTablesColumnIndexTransformer, self)._render_column_level(metric_column, display_schema)

        # Iterate through the pivoted levels
        i = 2 if display_schema.get('references') else 1
        data_keys, levels = [], []
        for dimension in list(display_schema['dimensions'].values())[1:]:
            level_key = metric_column[i]

            if 'display_options' in dimension:
                level_display = dimension['display_options'].get(level_key, level_key)
            else:
                if 'display_field' in dimension:
                    i += 1

                level_display = metric_column[i]

            # the metric key must remain last
            if not (isinstance(level_key, float) and np.isnan(level_key)):
                if level_key != Totals.key:
                    levels.append(level_display)
                data_keys.append(level_key)

            i += 1

        if levels:
            metric_label = '{metric} ({levels})'.format(
                metric=column['title'],
                levels=', '.join(map(str, levels))
            )
        else:
            metric_label = column['title']

        if data_keys and '.' in column['data']:
            # Squeeze the dimension levels between the reference and the metric
            data = column['data'].replace('.', '.' + ('.'.join(map(str, data_keys))) + '.')
        else:
            # Prepend the dimension levels to the metric
            data = '.'.join(map(str, data_keys + [column['data']]))

        return {
            'title': metric_label,
            'data': data,
            'render': {'type': 'value', '_': 'display', 'sort': 'value'}
        }

    def _recurse_dimensions(self, dataframe, dimensions, metrics, reference=None):
        if dataframe.empty:
            return []

        if reference or not dimensions:
            return super(DataTablesColumnIndexTransformer, self)._recurse_dimensions(dataframe, dimensions, metrics,
                                                                                     reference)

        data = []

        if 'display_field' in dimensions[0][1]:
            # Returning indexes per level does not guarantee they are ordered, therefore iteritems() must be used
            levels = [row[0][1:] for row in dataframe.iteritems()]
            format_key = lambda level: level[0]
            slice_metric_data = lambda level: dataframe[:, level[0], level[1]]
        else:
            levels = dataframe.index.levels[1]
            format_key = str
            slice_metric_data = lambda level: dataframe[:, level]

        for level in levels:
            metric_data = dict(self._recurse_dimensions(slice_metric_data(level), dimensions[1:], metrics))

            if not metric_data:
                continue

            data.append((format_key(level), metric_data))

        return data


class CSVRowIndexTransformer(DataTablesRowIndexTransformer):
    def transform(self, dataframe, display_schema):
        csv_df = self._format_columns(dataframe, display_schema['metrics'], display_schema['dimensions'])

        if isinstance(dataframe.index, pd.RangeIndex):
            # If there are no dimensions, just serialize to csv without the index
            return csv_df.to_csv(index=False)

        csv_df = self._format_index(csv_df, display_schema['dimensions'])

        row_dimension_labels = self._format_row_dimension_labels(display_schema['dimensions'])
        return csv_df.to_csv(index_label=row_dimension_labels)

    def _format_index(self, csv_df, dimensions):
        levels = list(dimensions.items())[:None if isinstance(csv_df.index, pd.MultiIndex) else 1]

        csv_df.index = [self.get_level_values(csv_df, key, dimension)
                        for key, dimension in levels]
        csv_df.index.names = [key
                              for key, dimension in levels]
        return csv_df

    def get_level_values(self, csv_df, key, dimension):
        if 'display_options' in dimension:
            return [_safe(dimension['display_options'].get(value, value))
                    for value in csv_df.index.get_level_values(key)]

        if 'display_field' in dimension:
            return [_safe(data_point)
                    for data_point in csv_df.index.get_level_values(dimension['display_field'])]

        return [_safe(data_point)
                for data_point in csv_df.index.get_level_values(key)]

    @staticmethod
    def _format_dimension_label(idx, dim_ordinal, dimension):
        if 'display_field' in dimension:
            display_field = dimension['display_field']
            return idx[dim_ordinal[display_field]]

        if isinstance(idx, tuple):
            id_field = dimension['definition'][0]
            dimension_display = idx[dim_ordinal[id_field]]

        else:
            dimension_display = idx

        if 'display_options' in dimension:
            return dimension['display_options'].get(dimension_display, dimension_display)

        return dimension_display

    def _format_columns(self, dataframe, metrics, dimensions):
        return dataframe.rename(columns=lambda metric: metrics[metric].get('label', metric))

    def _format_row_dimension_labels(self, dimensions):
        return [dimension['label']
                for dimension in dimensions.values()]


class CSVColumnIndexTransformer(DataTablesColumnIndexTransformer, CSVRowIndexTransformer):
    def _format_columns(self, dataframe, metrics, dimensions):
        if 1 < len(dimensions):
            csv_df = self._prepare_dataframe(dataframe, dimensions)
            csv_df.columns = _format_column_display(csv_df, metrics, dimensions)
            return csv_df

        return super(CSVColumnIndexTransformer, self)._format_columns(dataframe, metrics, dimensions)

    def _format_row_dimension_labels(self, dimensions):
        return [dimension['label']
                for dimension in list(dimensions.values())[:1]]



# coding: utf-8
from fireant.dashboards.managers import WidgetGroupManager

from fireant.slicer.transformers import *


class Widget(object):
    """
    The `Widget` class represents a chart or table widget in one of the dashboard's sections.

    Attributes:
        * label: human readable text.
        * metrics: list of metric names/identifiers
    """

    def __init__(self, metrics=tuple()):
        self.metrics = metrics


class LineChartWidget(Widget):
    """
    The `LineChartWidget` class represents a Highcharts line chart.
    """
    transformer = HighchartsLineTransformer()


class AreaChartWidget(Widget):
    """
    The `AreaChartWidget` class represents a Highcharts area chart.
    """
    transformer = HighchartsAreaTransformer()


class AreaPercentageChartWidget(Widget):
    """
    The `AreaPercentageChartWidget` class represents a Highcharts area percentage chart.
    """
    transformer = HighchartsAreaPercentageTransformer()


class BarChartWidget(Widget):
    """
    The `BarChartWidget` class represents a Highcharts bar chart.
    """
    transformer = HighchartsBarTransformer()


class StackedBarChartWidget(Widget):
    """
    The `StackedBarChartWidget` class represents a Highcharts stacked bar chart.
    """
    transformer = HighchartsStackedBarTransformer()


class ColumnChartWidget(Widget):
    """
    The `ColumnChartWidget` class represents a Highcharts column chart.
    """
    transformer = HighchartsColumnTransformer()


class StackedColumnChartWidget(Widget):
    """
    The `StackedColumnChartWidget` class represents a Highcharts stacked column chart.
    """
    transformer = HighchartsStackedColumnTransformer()


class PieChartWidget(Widget):
    """
    The `PieChartWidget` class represents a Highcharts Pie chart.
    """
    transformer = HighchartsPieTransformer()


class RowIndexTableWidget(Widget):
    """
    The `RowIndexTableWidget` class represents datatables.js data table with row-indexed dimensions.
    """
    transformer = DataTablesRowIndexTransformer()


class ColumnIndexTableWidget(Widget):
    """
    The `ColumnIndexTableWidget` class represents datatables.js data table with column-indexed dimensions.
    """
    transformer = DataTablesColumnIndexTransformer()


class WidgetGroup(object):
    def __init__(self, slicer, widgets=None, dimensions=None, dimension_filters=None, references=None, operations=None):
        self.slicer = slicer
        self.widgets = widgets

        self.dimensions = dimensions or []
        self.dimension_filters = dimension_filters or []
        self.references = references or []
        self.operations = operations or []

        self.manager = WidgetGroupManager(self)



# coding: utf-8

from .schemas import (WidgetGroup, LineChartWidget,
                      AreaChartWidget, AreaPercentageChartWidget,
                      BarChartWidget, StackedBarChartWidget,
                      ColumnChartWidget, StackedColumnChartWidget,
                      PieChartWidget,
                      RowIndexTableWidget, ColumnIndexTableWidget)



# coding: utf-8

import functools
import itertools
from collections import OrderedDict

import numpy as np
import pandas as pd

from fireant import utils
from fireant.slicer.operations import Totals
from pypika import functions as fn
from .postprocessors import OperationManager
from .queries import QueryManager


class SlicerException(Exception):
    pass


class SlicerManager(QueryManager, OperationManager):
    def __init__(self, slicer):
        """
        :param slicer:
        """
        self.slicer = slicer

    def data(self, metrics=(), dimensions=(),
             metric_filters=(), dimension_filters=(),
             references=(), operations=()):
        """
        :param metrics:
            Type: list or tuple
            A set of metrics to include in the query.

        :param dimensions:
            Type: list or tuple
            A set of dimensions to split the metrics into groups.

        :param metric_filters:
            Type: list or tuple
            A set of filters to constrain the data with by metric thresholds.

        :param dimension_filters:
            Type: list or tuple
            A set of filters to constrain the data with by dimension.

        :param references:
            Type: list or tuple
            A set of comparisons to include in the query.

        :param operations:
            Type: list or tuple
            A set of operations to perform on the response.

        :return:
            A transformed response that is queried based on the slicer and the format.
        """
        metrics, dimensions = map(utils.filter_duplicates, (metrics, dimensions))

        query_schema = self.data_query_schema(metrics=metrics, dimensions=dimensions,
                                              metric_filters=metric_filters, dimension_filters=dimension_filters,
                                              references=references, operations=operations)
        operation_schema = self.operation_schema(operations)

        dataframe = self.query_data(**query_schema)
        dataframe = self.post_process(dataframe, operation_schema)

        # Filter additional metrics from the dataframe that were needed for operations
        final_columns = metrics + ['%s_%s' % (os['metric'], os['key']) for os in operation_schema]
        if not references:
            return dataframe[final_columns]

        reference_columns = [''] + [r.key for r in references]
        return dataframe[list(itertools.product(reference_columns, final_columns))]

    def query_string(self, metrics=(), dimensions=(),
                     metric_filters=(), dimension_filters=(),
                     references=(), operations=()):
        """
        :param metrics:
            Type: list or tuple
            A set of metrics to include in the query.

        :param dimensions:
            Type: list or tuple
            A set of dimensions to split the metrics into groups.

        :param metric_filters:
            Type: list or tuple
            A set of filters to constrain the data with by metric thresholds.

        :param dimension_filters:
            Type: list or tuple
            A set of filters to constrain the data with by dimension.

        :param references:
            Type: list or tuple
            A set of comparisons to include in the query.

        :param operations:
            Type: list or tuple
            A set of operations to perform on the response.

        :return:
            The query that would generate the data.
        """
        metrics = utils.filter_duplicates(metrics)
        dimensions = utils.filter_duplicates(dimensions)

        query_schema = self.data_query_schema(metrics=metrics, dimensions=dimensions,
                                              metric_filters=metric_filters, dimension_filters=dimension_filters,
                                              references=references, operations=operations)

        return str(self._build_data_query(**query_schema))

    def dimension_options(self, dimension, filters, limit=None):
        dimopt_schema = self.dimension_option_schema(dimension, filters, limit)
        return self.query_dimension_options(**dimopt_schema)

    def data_query_schema(self, metrics=(), dimensions=(),
                          metric_filters=(), dimension_filters=(),
                          references=(), operations=()):
        """
        Builds a `dict` model of the schema parts required for executing a data query given a request.

        :param metrics:
        :param dimensions:
        :param metric_filters:
        :param dimension_filters:
        :param references:
        :param operations:

        :return:
        """
        metric_joins_schema = self._joins_schema(set(metrics) | {mf.element_key for mf in metric_filters},
                                                 self.slicer.metrics)
        dimension_joins_schema = self._joins_schema(set(dimensions) | {df.element_key for df in dimension_filters},
                                                    self.slicer.dimensions)
        return {
            'database': self.slicer.database,
            'table': self.slicer.table,

            'metrics': self._metrics_schema(metrics, operations),
            'dimensions': self._dimensions_schema(dimensions),

            'mfilters': self._filters_schema(self.slicer.metrics, metric_filters,
                                             self._default_metric_definition,
                                             element_label='metric'),
            'dfilters': self._filters_schema(self.slicer.dimensions,
                                             dimension_filters,
                                             self._default_dimension_definition),

            'joins': list(metric_joins_schema | dimension_joins_schema),
            'references': self._references_schema(references),
            'rollup': self._totals_schema(dimensions, operations),
        }

    def dimension_option_schema(self, dimension, filters, limit=None):
        dimensions = [dimension]

        schema_dimensions = self._dimensions_schema(dimensions)
        schema_filters = self._filters_schema(self.slicer.dimensions, filters, self._default_dimension_definition)
        schema_joins = self._joins_schema(set(dimensions) | {df.element_key for df in filters},
                                          self.slicer.dimensions)

        return {
            'database': self.slicer.database,
            'table': self.slicer.hint_table or self.slicer.table,
            'joins': schema_joins,
            'dimensions': schema_dimensions,
            'filters': schema_filters,
            'limit': limit,
        }

    def display_schema(self, metrics=(), dimensions=(), references=(), operations=()):
        """
        Builds a display schema for a request.  This is used in combination with the results of the query that is
        executed by the Slicer manager to transform the results with display labels.  The display schema carries
        meta-data pertaining to displaying the results of the query.

        :param metrics:
            Type: list[str]
            The requested list of metrics.  This must match a Metric contained in the slicer.
        :param dimensions:
            Type: list[str or tuple/list]
            The requested list of dimensions.  For simple dimensions, the string key is passed as a parameter, otherwise
            a tuple of list with the first element equal to the key and the subsequent elements equal to the paramters.
        :param references:
            Type: list[tuple]
            A list of tuples describing reference options, where the first element in the tuple is the key of the
            reference operation and the second value is the dimension to perform the comparasion along.
        :return:
            A dictionary describing how to transform the resulting data frame for the same request.
        """
        return {
            'metrics': self._display_metrics(metrics, operations),
            'dimensions': self._display_dimensions(dimensions, operations),
            'references': OrderedDict([(reference.key, reference.label)
                                       for reference in references]),
        }

    def operation_schema(self, operations):
        results = []
        for operation in operations:
            schema = operation.schemas()

            if schema is not None:
                results.append(schema)

        return results

    def _metrics_schema(self, metrics=(), operations=()):
        keys = list(metrics) + [metric
                                for op in operations
                                for metric in op.metrics()]

        if not keys:
            raise SlicerException('At least one metric is required requests.  Please add a metric.')

        invalid_metrics = {utils.slice_first(key)
                           for key in keys} - set(self.slicer.metrics)
        if invalid_metrics:
            raise SlicerException('Invalid metrics included in request: '
                                  '[%s]' % ', '.join(invalid_metrics))

        schema_metrics = OrderedDict()
        for key in keys:
            schema_metric = self.slicer.metrics.get(key)

            for metric_key, definition in schema_metric.schemas():
                schema_metrics[metric_key] = definition or self._default_metric_definition(metric_key)

        return schema_metrics

    def _dimensions_schema(self, keys):
        invalid_dimensions = {utils.slice_first(key) for key in keys} - set(self.slicer.dimensions)
        if invalid_dimensions:
            raise SlicerException('Invalid dimensions included in request: '
                                  '[%s]' % ', '.join(invalid_dimensions))

        dimensions = OrderedDict()
        for dimension in keys:
            # unpack tuples for args
            if isinstance(dimension, (list, tuple)):
                dimension, args = dimension[0], dimension[1:]
            else:
                args = []

            schema_dimension = self.slicer.dimensions.get(dimension)

            for key, definition in schema_dimension.schemas(*args, database=self.slicer.database):
                dimensions[key] = definition or self._default_dimension_definition(key)

        return dimensions

    def _joins_schema(self, keys, elements):
        """

        :param keys:
            The keys of the schema elements to retrieve joins for.
        :param elements:
            The elements to retrieve the joins from, either slicer.metrics or slicer.dimensions.
        :return:
            A `set` of join schemas containing the join table and the join criterion.
        """
        joins = set()

        for key in keys:
            element = elements.get(utils.slice_first(key))
            if element and element.joins:
                joins |= set(element.joins)

        return {(self.slicer.joins[key].table,
                 self.slicer.joins[key].criterion,
                 self.slicer.joins[key].join_type) for key in joins}

    def _filters_schema(self, elements, filters, default_value_func, element_label='dimension'):
        filters_schema = []
        for filter_item in filters:
            if isinstance(filter_item.element_key, (tuple, list)):
                element_key, modifier = filter_item.element_key
            else:
                element_key, modifier = filter_item.element_key, None

            element = elements.get(element_key)
            if not element:
                raise SlicerException(
                    'Unable to apply filter [{filter}].  '
                    'No such {element} with key [{key}].'.format(
                        filter=filter_item,
                        element=element_label,
                        key=filter_item.element_key
                    ))

            if hasattr(element, 'display_field') and 'display' == modifier:
                definition = element.display_field

            else:
                definition = element.definition or default_value_func(element.key)

            filters_schema.append(filter_item.schemas(definition))

        return filters_schema

    def _display_dimensions(self, dimensions, operations):
        req_dimension_keys = [utils.slice_first(dimension)
                              for dimension in dimensions]

        display_dims = OrderedDict()
        for key in req_dimension_keys:
            dimension = self.slicer.dimensions[key]
            display_dim = {'label': dimension.label}

            if hasattr(dimension, 'display_options'):
                display_dim['display_options'] = {opt.key: opt.label
                                                  for opt in dimension.display_options}
                display_dim['display_options'].update({pd.NaT: '', np.nan: ''})

            if hasattr(dimension, 'display_field') and dimension.display_field:
                display_dim['display_field'] = '%s_display' % dimension.key

            display_dims[key] = display_dim

        return display_dims

    def _references_schema(self, references):
        schema_references = OrderedDict()
        for reference in references:
            if reference.element_key not in self.slicer.dimensions.keys():
                raise SlicerException(
                    'Unable to query with [{reference}]. '
                    'No such dimension [{dimension}].'.format(reference=str(reference),
                                                              dimension=reference.element_key))

            from .schemas import DatetimeDimension
            if not isinstance(self.slicer.dimensions[reference.element_key], DatetimeDimension):
                raise SlicerException(
                    'Unable to query with [{reference}]. '
                    'Dimension [{dimension}] must be a DatetimeDimension.'.format(reference=str(reference),
                                                                                  dimension=reference.element_key))

            schema_references[reference.key] = {
                'dimension': reference.element_key,
                'definition': self.slicer.dimensions[reference.element_key].definition,
                'interval': reference.interval,
                'modifier': reference.modifier,
            }

        return schema_references

    def _default_dimension_definition(self, key):
        return fn.Coalesce(self.slicer.table.field(key), 'None')

    def _default_metric_definition(self, key):
        return fn.Sum(self.slicer.table.field(key))

    def _display_metrics(self, metrics, operations):
        display = OrderedDict()

        axis = 0
        for metrics_level in metrics:
            for metric_key in utils.wrap_list(metrics_level):
                schema = self.slicer.metrics[metric_key]
                display[metric_key] = {attr: getattr(schema, attr)
                                       for attr in ['label', 'precision', 'prefix', 'suffix']
                                       if getattr(schema, attr) is not None}
                display[metric_key]["axis"] = axis
            axis += 1

        for operation in operations:
            if not hasattr(operation, 'metric_key'):
                continue

            metric_key = operation.metric_key
            metric_schema = self.slicer.metrics[metric_key]

            key = '{}_{}'.format(metric_key, operation.key)
            display[key] = {attr: getattr(metric_schema, attr)
                            for attr in ['precision', 'prefix', 'suffix']
                            if getattr(metric_schema, attr) is not None}
            display[key]['label'] = '{} {}'.format(metric_schema.label, operation.label)
            display[key]["axis"] = axis
            axis += 1

        return display

    def _totals_schema(self, dimensions, operations):
        dimension_set = set(utils.slice_first(dimension) for dimension in dimensions)
        totals, missing_dimensions = [], set()

        for operation in operations:
            if operation.key != Totals.key:
                continue

            missing_dimensions |= set(operation.dimension_keys) - dimension_set
            totals += [[level for level in self.slicer.dimensions[dimension].levels()]
                       for dimension in operation.dimension_keys
            ]

        if missing_dimensions:
            raise SlicerException("Missing dimensions with keys: {}".format(", ".join(missing_dimensions)))

        return totals


class TransformerManager(object):
    def __init__(self, manager, transformers):
        self.manager = manager

        # Creates a function on the slicer for each transformer
        for tx_key, tx in transformers.items():
            setattr(self, tx_key, functools.partial(self._get_and_transform_data, tx))

    def _get_and_transform_data(self, tx, metrics=(), dimensions=(),
                                metric_filters=(), dimension_filters=(),
                                references=(), operations=()):
        """
        Handles a request and applies a transformation to the result.  This is the implementation of all of the
        transformer manager methods, which are constructed in the __init__ function of this class for each transformer.

        The request is first validated with the transformer then the request is executed via the SlicerManager and then
        lastly the result is transformed and returned.

        :param tx:
            The transformer to use
        :param metrics:
            See ``fireant.slicer.SlicerManager``
            A list of metrics to include in the query.

        :param dimensions:
            See ``fireant.slicer.SlicerManager``
            A list of dimensions to include in the query.

        :param metric_filters:
            See ``fireant.slicer.SlicerManager``
            A list of metrics filters to apply to the query.

        :param dimension_filters:
            See ``fireant.slicer.SlicerManager``
            A list of dimension filters to apply to the query.

        :param references:
            See ``fireant.slicer.SlicerManager``
            A list of references to include in the query

        :param operations:
            See ``fireant.slicer.SlicerManager``
            A list of post-operations to apply to the result before transformation.

        :return:
            The transformed result of the request.
        """
        tx.prevalidate_request(self.manager.slicer, metrics=metrics,
                               dimensions=[utils.slice_first(dimension)
                                           for dimension in dimensions],
                               metric_filters=metric_filters, dimension_filters=dimension_filters,
                               references=references, operations=operations)

        # Loads data and transforms it with a given transformer.
        dataframe = self.manager.data(metrics=utils.flatten(metrics), dimensions=dimensions,
                                      metric_filters=metric_filters, dimension_filters=dimension_filters,
                                      references=references, operations=operations)
        display_schema = self.manager.display_schema(metrics, dimensions, references, operations)

        return tx.transform(dataframe, display_schema)



# coding: utf-8

from .base import (Transformer,
                   TransformationException)
from .datatables import (DataTablesRowIndexTransformer,
                         DataTablesColumnIndexTransformer)
from .datatables import (DataTablesRowIndexTransformer,
                         DataTablesColumnIndexTransformer,
                         CSVRowIndexTransformer,
                         CSVColumnIndexTransformer)
from .highcharts import (HighchartsLineTransformer,
                         HighchartsAreaTransformer,
                         HighchartsAreaPercentageTransformer,
                         HighchartsColumnTransformer,
                         HighchartsStackedColumnTransformer,
                         HighchartsBarTransformer,
                         HighchartsStackedBarTransformer,
                         HighchartsPieTransformer)
from .notebooks import (PandasRowIndexTransformer,
                        PandasColumnIndexTransformer,
                        MatplotlibLineChartTransformer,
                        MatplotlibBarChartTransformer)

BUNDLES = {
    'notebooks': {
        'row_index_table': PandasRowIndexTransformer(),
        'column_index_table': PandasColumnIndexTransformer(),
        'line_chart': MatplotlibLineChartTransformer(),
        'bar_chart': MatplotlibBarChartTransformer(),
    },
    'highcharts': {
        'line_chart': HighchartsLineTransformer(),
        'area_chart': HighchartsAreaTransformer(),
        'area_percentage_chart': HighchartsAreaPercentageTransformer(),
        'column_chart': HighchartsColumnTransformer(),
        'stacked_column_chart': HighchartsStackedColumnTransformer(),
        'bar_chart': HighchartsBarTransformer(),
        'stacked_bar_chart': HighchartsStackedBarTransformer(),
        'pie_chart': HighchartsPieTransformer(),
    },
    'datatables': {
        'row_index_table': DataTablesRowIndexTransformer(),
        'column_index_table': DataTablesColumnIndexTransformer(),
        'row_index_csv': CSVRowIndexTransformer(),
        'column_index_csv': CSVColumnIndexTransformer(),
    },
}



# coding: utf-8

import pandas as pd


class Database(object):
    def connect(self):
        raise NotImplementedError

    def trunc_date(self, field, interval):
        raise NotImplementedError

    def interval(self, years=0, months=0, days=0, hours=0, minutes=0, seconds=0, microseconds=0, quarters=0, weeks=0):
        raise NotImplementedError

    def fetch(self, query):
        with self.connect() as connection:
            cursor = connection.cursor()
            cursor.execute(query)
            return cursor.fetchall()

    def fetch_dataframe(self, query):
        with self.connect() as connection:
            return pd.read_sql(query, connection)



# coding: utf-8
from pypika import Interval, functions as fn


class Reference(object):
    key = None
    label = None
    interval = None
    modifier = None

    def __init__(self, element_key):
        self.element_key = element_key

    def __eq__(self, other):
        if not isinstance(other, Reference):
            return False

        return all([
            self.key == other.key,
            self.element_key == other.element_key,
        ])

    def __hash__(self):
        return hash('%s:%s' % (self.key, self.element_key))


class Delta(Reference):
    def __init__(self, reference):
        super(Delta, self).__init__(reference.element_key)
        self.key = self.generate_key(reference.key)
        self.label = '%s Δ' % reference.label
        self.interval = reference.interval

    @staticmethod
    def modifier(field, join_field):
        return field - join_field

    @staticmethod
    def generate_key(reference_key):
        return '%s_delta' % reference_key


class DeltaPercentage(Reference):
    def __init__(self, reference):
        super(DeltaPercentage, self).__init__(reference.element_key)
        self.key = self.generate_key(reference.key)
        self.label = '%s Δ%%' % reference.label
        self.interval = reference.interval

    @staticmethod
    def modifier(field, join_field):
        return ((field - join_field) * 100) / fn.NullIf(join_field, 0)

    @staticmethod
    def generate_key(reference_key):
        return '%s_delta_percent' % reference_key


class DoD(Reference):
    key = 'dod'
    label = 'DoD'
    interval = Interval(days=1)


class WoW(Reference):
    key = 'wow'
    label = 'WoW'
    interval = Interval(weeks=1)


class MoM(Reference):
    key = 'mom'
    label = 'MoM'
    interval = Interval(months=1)


class QoQ(Reference):
    key = 'qoq'
    label = 'QoQ'
    interval = Interval(quarters=1)


class YoY(Reference):
    key = 'yoy'
    label = 'YoY'
    interval = Interval(years=1)



# coding: utf-8
import pandas as pd

from fireant import utils
from fireant.slicer.managers import Totals


class WidgetGroupManager(object):
    def __init__(self, widget_group):
        self.widget_group = widget_group

    def _schema(self, dimensions=None, metric_filters=None, dimension_filters=None, references=None, operations=None):
        return dict(
            metrics=[metric
                     for widget in self.widget_group.widgets
                     for metric in utils.flatten(widget.metrics)],
            dimensions=dimensions,
            metric_filters=metric_filters or [],
            dimension_filters=self.widget_group.dimension_filters + (dimension_filters or []),
            references=references,
            operations=operations,
        )

    def render(self, dimensions=None, metric_filters=None, dimension_filters=None, references=None, operations=None):
        dimensions = utils.filter_duplicates(self.widget_group.dimensions + (dimensions or []))
        references = utils.filter_duplicates(self.widget_group.references + (references or []))
        operations = utils.filter_duplicates(self.widget_group.operations + (operations or []))

        for widget in self.widget_group.widgets:
            widget.transformer.prevalidate_request(self.widget_group.slicer, widget.metrics, dimensions,
                                                   metric_filters, dimension_filters, references, operations)

        schema = self._schema(dimensions, metric_filters, dimension_filters, references, operations)
        dataframe = self.widget_group.slicer.manager.data(**schema)

        _args = [dataframe, schema['dimensions'], schema['references'], schema['operations']]
        return [self._transform_widget(widget, *_args)
                for widget in self.widget_group.widgets]

    def query_string(self, dimensions=None, metric_filters=None, dimension_filters=None, references=None,
                     operations=None):
        dimensions = utils.filter_duplicates(self.widget_group.dimensions + (dimensions or []))
        references = utils.filter_duplicates(self.widget_group.references + (references or []))
        operations = utils.filter_duplicates(self.widget_group.operations + (operations or []))

        schema = self._schema(dimensions, metric_filters, dimension_filters, references, operations)
        return self.widget_group.slicer.manager.query_string(**schema)

    def _transform_widget(self, widget, dataframe, dimensions, references, operations):
        display_schema = self.widget_group.slicer.manager.display_schema(
            metrics=widget.metrics,
            dimensions=dimensions,
            references=references,
            operations=operations,
        )

        # Temporary fix to enable operations to get output properly. Can removed when the Fireant API is refactored.
        operation_columns = ['{}_{}'.format(operation.metric_key, operation.key)
                             for operation in operations if operation.key != Totals.key]

        columns = utils.flatten(widget.metrics) + operation_columns

        if references:
            # This escapes a pandas bug where a data frame subset of columns still returns the columns of the
            # original data frame
            reference_keys = [''] + [ref.key for ref in references]
            subset_columns = pd.MultiIndex.from_product([reference_keys, columns])
            subset = pd.DataFrame(dataframe[subset_columns], columns=subset_columns)

        else:
            subset = dataframe[columns]

        return widget.transformer.transform(subset, display_schema)



# coding: utf8

from fireant import __version__
from setuptools import setup

setup(
    name='fireant',
    version=__version__,

    author='KAYAK, GmbH',
    author_email='bandit@kayak.com',

    # License
    license='Apache License Version 2.0',

    packages=['fireant', 'fireant.database', 'fireant.dashboards', 'fireant.slicer', 'fireant.slicer.transformers'],

    # Include additional files into the package
    include_package_data=True,

    # Details
    url='https://github.com/kayak/fireant',

    description='Data analysis tool for Python and Jupyter Notebooks',
    long_description="Fireant is a tool which aims to simplify the process of accessing analytical data and building "
                     "dashboards and reports.  It eliminates the need for handwritten queries or code to process data. "
                     "The main feature, the slicer, is a schema that is used to configure metrics, the data being "
                     "visualized, and dimensions, the spatiality over which metrics can be extended and grouped. "
                     "Included in the Slicer is a Slicer Manager API which exposes methods to render tables and charts "
                     "in a variety of different formats.\n\n"

                     "Fireant is a great compliment to Jupter notebooks using matplotlib and pandas. It can be used in "
                     "web projects with Highcharts and Datatables.js.  It can even just provide a layer of abstraction "
                     "for accessing analytical data akin to ORM libraries in software projects.",

    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: PL/SQL',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ],
    keywords=('fireant python query builder querybuilder sql mysql postgres psql oracle vertica aggregated '
              'relational database rdbms business analytics bi data science analysis pandas'),

    install_requires=[
        'six',
        'pandas',
        'pypika',
        'vertica-python>=0.6'
    ],
    tests_require=[
        'mock'
    ],
    extras_require={
        'vertica': ['vertica-python>=0.6'],
        'matplotlib': ['matplotlib'],
    },

    test_suite='fireant.tests',
)



# coding: utf-8
import numpy as np
import pandas as pd

from collections import Counter

from fireant import settings, utils
from fireant.slicer.operations import Totals
from .base import Transformer, TransformationException

COLORS = {
    'kayak': ['#FF690F', '#3083F0', '#00B86B', '#D10244', '#FFDBE5', '#7B4F4B', '#B903AA', '#B05B6F'],
    'dark-blue': ["#DDDF0D", "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                  "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'dark-green': ["#DDDF0D", "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                   "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'dark-unica': ["#2B908F", "#90EE7E", "#F45B5B", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                   "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'gray': ["#DDDF0D", "#7798BF", "#55BF3B", "#DF5353", "#AAEEEE", "#FF0066", "#EEAAEE",
             "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'grid-light': ["#7CB5EC", "#F7A35C", "#90EE7E", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                   "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'grid': ['#058DC7', '#50B432', '#ED561B', '#DDDF00', '#24CBE5', '#64E572', '#FF9655', '#FFF263', '#6AF9C4'],
    'sand-signika': ["#F45B5B", "#8085E9", "#8D4654", "#7798BF", "#AAEEEE", "#FF0066", "#EEAAEE",
                     "#55BF3B", "#DF5353", "#7798BF", "#AAEEEE"],
    'skies': ["#514F78", "#42A07B", "#9B5E4A", "#72727F", "#1F949A", "#82914E", "#86777F", "#42A07B"],
}


def _format_data_point(value):
    if isinstance(value, pd.Timestamp):
        return int(value.asm8) // int(1e6)
    if value is None or (isinstance(value, (float, int)) and np.isnan(value)):
        return None
    if isinstance(value, np.int64):
        # Cannot serialize np.int64 to json
        return int(value)
    return value


def _color(i):
    colors = COLORS.get(settings.highcharts_colors, 'grid')
    n_colors = len(colors)

    return colors[i % n_colors]


MISSING_CONT_DIM_MESSAGE = ('Highcharts line charts require a continuous dimension as the first '
                            'dimension.  Please add a continuous dimension from your Slicer to '
                            'your request.')


class HighchartsLineTransformer(Transformer):
    """
    Transforms data frames into Highcharts format for several chart types, particularly line or bar charts.
    """

    chart_type = 'line'

    def prevalidate_request(self, slicer, metrics, dimensions,
                            metric_filters, dimension_filters,
                            references, operations):
        if not dimensions or not slicer.dimensions:
            raise TransformationException(MISSING_CONT_DIM_MESSAGE)

        from fireant.slicer import ContinuousDimension
        first_dimension = utils.slice_first(dimensions[0])
        dimension0 = slicer.dimensions[first_dimension]
        if not isinstance(dimension0, ContinuousDimension):
            raise TransformationException(MISSING_CONT_DIM_MESSAGE)

    def transform(self, dataframe, display_schema):
        has_references = isinstance(dataframe.columns, pd.MultiIndex)

        dim_ordinal = {name: ordinal
                       for ordinal, name in enumerate(dataframe.index.names)}
        dataframe = self._prepare_dataframe(dataframe, dim_ordinal, display_schema['dimensions'])

        if has_references:
            series = sum(
                [self._make_series(dataframe[level], dim_ordinal, display_schema, reference=level or None)
                 for level in dataframe.columns.levels[0]],
                []
            )

        else:
            series = self._make_series(dataframe, dim_ordinal, display_schema)

        result = {
            'chart': {'type': self.chart_type, 'zoomType': 'x'},
            'title': {'text': None},
            'plotOptions': {},
            'xAxis': self.xaxis_options(dataframe, dim_ordinal, display_schema),
            'yAxis': self.yaxis_options(dataframe, dim_ordinal, display_schema),
            'tooltip': {'shared': True},
            'series': series
        }

        return result

    def xaxis_options(self, dataframe, dim_ordinal, display_schema):
        return {
            'type': 'datetime' if isinstance(dataframe.index, pd.DatetimeIndex) else 'linear'
        }

    def yaxis_options(self, dataframe, dim_ordinal, display_schema):
        metrics_per_axis = Counter((metric['axis'] for metric in display_schema['metrics'].values()))

        metric_idx = 0
        axis_options = {}

        for axes_id, metrics_count in metrics_per_axis.items():
            metric_idx += metrics_count
            axis_options[axes_id] = {
                'opposite': metrics_count > 1, 'color': 'black' if metrics_count > 1 else _color(metric_idx - 1)
            }

        # Axes with just one metric have the same color of their line. References and multiple metrics' axis
        # will default to black.
        axis = [
            {
                'id': axes_id, 'opposite': options['opposite'], 'title': {'text': None},
                'labels': {
                    'style': {
                        'color': options['color']
                    }
                }
            }
            for axes_id, options in axis_options.items()
        ]

        reference_keys = display_schema.get('references', {}).keys()

        # Create only one axes per available modifier (i.e. delta percent, percent and none).
        reference_axes_ids = set(self._reference_axes_id(reference_key) for reference_key in reference_keys)

        for i, axes_id in enumerate(reference_axes_ids):
            axis.append({'id': axes_id, 'title': {'text': 'References' if i == 0 else None}, 'opposite': True})

        return axis


    def _make_series(self, dataframe, dim_ordinal, display_schema, reference=None):
        metrics = list(dataframe.columns.levels[0]
                       if isinstance(dataframe.columns, pd.MultiIndex)
                       else dataframe.columns)

        return [self._make_series_item(idx, item, dim_ordinal, display_schema, metrics, reference, _color(i))
                for i, (idx, item) in enumerate(dataframe.iteritems())]

    def _make_series_item(self, idx, item, dim_ordinal, display_schema, metrics, reference, color='#000'):
        metric_key = utils.slice_first(idx)

        return {
            'name': self._format_label(idx, dim_ordinal, display_schema, reference),
            'data': self._format_data(item),
            'tooltip': self._format_tooltip(display_schema['metrics'][metric_key]),
            'yAxis': display_schema['metrics'][metric_key].get('axis', 0)
                     if not reference else self._reference_axes_id(reference),
            'color': color,
            'dashStyle': 'Dot' if reference else 'Solid'
        }

    @staticmethod
    def _reference_axes_id(reference_key):
        if '_' in reference_key:
            # Replace dod, wow, mom and yoy with reference, since each modifier type should have only one axes.
            modifier = '_'.join(reference_key.split('_')[1:])
            return 'reference_{}'.format(modifier)
        return 'reference'

    @staticmethod
    def _make_categories(dataframe, dim_ordinal, display_schema):
        return None

    def _format_tooltip(self, metric_schema):
        tooltip = {}

        if 'precision' in metric_schema:
            tooltip['valueDecimals'] = metric_schema['precision']
        if 'prefix' in metric_schema:
            tooltip['valuePrefix'] = metric_schema['prefix']
        if 'suffix' in metric_schema:
            tooltip['valueSuffix'] = metric_schema['suffix']

        return tooltip

    def _prepare_dataframe(self, dataframe, dim_ordinal, dimensions):
        # Replaces invalid values and unstacks the data frame for line charts.

        # Force all fields to be float (Safer for highcharts)
        dataframe = dataframe.astype(np.float).replace([np.inf, -np.inf], np.nan)

        # Unstack multi-indices
        if 1 < len(dimensions):
            # We need to unstack all of the dimensions here after the first dimension, which is the first dimension in
            # the dimensions list, not necessarily the one in the dataframe
            unstack_levels = list(self._unstack_levels(list(dimensions.items())[1:], dim_ordinal))
            dataframe = dataframe.unstack(level=unstack_levels)

        return dataframe

    def _format_label(self, idx, dim_ordinal, display_schema, reference):
        is_multidimensional = isinstance(idx, tuple)

        metric_idx = idx[0] if is_multidimensional else idx
        metric = display_schema['metrics'].get(metric_idx, {})
        metric_label = metric.get('label', metric_idx)

        if reference:
            metric_label += ' {}'.format(display_schema['references'][reference])

        if not is_multidimensional:
            return metric_label

        dimension_labels = [self._format_dimension_display(dim_ordinal, key, dimension, idx)
                            for key, dimension in list(display_schema['dimensions'].items())[1:]]

        dimension_labels = [dimension_label  # filter out the Totals
                            for dimension_label in dimension_labels
                            if dimension_label and dimension_label is not Totals.label]

        return (
            '{metric} ({dimensions})'.format(
                metric=metric_label,
                dimensions=', '.join(map(str, dimension_labels))
            )
            if dimension_labels
            else metric_label
        )

    @staticmethod
    def _format_dimension_display(dim_ordinal, key, dimension, idx):
        if not isinstance(idx, (list, tuple)):
            idx = [idx]

        if 'display_field' in dimension:
            display_field = dimension['display_field']
            return idx[dim_ordinal[display_field]]

        dimension_value = idx[dim_ordinal[key]]

        if 'display_options' in dimension:
            dimension_value = dimension['display_options'].get(dimension_value, dimension_value)

        return dimension_value

    def _format_data(self, column):
        if isinstance(column, float):
            return [_format_data_point(column)]

        return [self._format_point(key, value)
                for key, value in column.iteritems()
                if not (isinstance(value, (float, int)) and np.isnan(value)) and not pd.isnull(key)]

    @staticmethod
    def _format_point(x, y):
        return (_format_data_point(x), _format_data_point(y))

    def _unstack_levels(self, dimensions, dim_ordinal):
        for key, dimension in dimensions:
            yield dim_ordinal[key]

            if 'display_field' in dimension:
                yield dim_ordinal[dimension['display_field']]


class HighchartsAreaTransformer(HighchartsLineTransformer):
    """
    Transformer for a Highcharts Area chart
    http://www.highcharts.com/demo/area-basic
    """
    chart_type = 'area'

    def _make_series_item(self, idx, item, dim_ordinal, display_schema, metrics, reference, color='#000'):
        """
        Overriding the parent class' _make_series_item to remove the yAxis key as area charts
        only really make sense on a single y axis
        """
        metric_key = utils.slice_first(idx)
        return {
            'name': self._format_label(idx, dim_ordinal, display_schema, reference),
            'data': self._format_data(item),
            'tooltip': self._format_tooltip(display_schema['metrics'][metric_key]),
            'color': color,
            'dashStyle': 'Dot' if reference else 'Solid'
        }

    def yaxis_options(self, dataframe, dim_ordinal, display_schema):
        return [{'title': None}]


class HighchartsAreaPercentageTransformer(HighchartsAreaTransformer):
    """
    Transformer for a Highcharts Area Percentage chart
    http://www.highcharts.com/demo/area-stacked-percent
    """
    chart_type = 'area'

    def transform(self, dataframe, display_schema):
        config = super(HighchartsAreaPercentageTransformer, self).transform(dataframe, display_schema)
        config['plotOptions'] = {
            'area': {
                'stacking': 'percent',
            }
        }
        return config

    def _format_tooltip(self, metric_schema):
        tooltip = super(HighchartsAreaPercentageTransformer, self)._format_tooltip(metric_schema)
        # Add the percentage to the default tooltip point format
        tooltip['pointFormat'] = '<span style="color:{point.color}">\u25CF</span> {series.name}: ' \
                                 '<b>{point.y} - {point.percentage:.1f}%</b><br/>'
        return tooltip


class HighchartsColumnTransformer(HighchartsLineTransformer):
    """
    Transformer for a Highcharts Column chart
    http://www.highcharts.com/demo/column-basic
    """
    chart_type = 'column'

    def prevalidate_request(self, slicer, metrics, dimensions,
                            metric_filters, dimension_filters,
                            references, operations):
        if dimensions and 2 < len(dimensions):
            # Too many dimensions
            raise TransformationException('Highcharts bar and column charts support at a maximum two dimensions.  '
                                          'Request included %d dimensions.' % len(dimensions))

        if dimensions and 2 == len(dimensions) and metrics and 1 < len(metrics):
            # Too many metrics
            raise TransformationException('Highcharts bar and column charts support at a maximum one metric with '
                                          'two dimensions.  Please remove some dimensions or metrics.  '
                                          'Request included %d metrics and %d dimensions.' % (len(metrics),
                                                                                              len(dimensions)))

    def _make_series_item(self, idx, item, dim_ordinal, display_schema, metrics, reference, color='#000'):
        metric_key = utils.slice_first(idx)
        return {
            'name': self._format_label(idx, dim_ordinal, display_schema, reference),
            'data': self._format_data(item),
            'tooltip': self._format_tooltip(display_schema['metrics'][metric_key]),
            'yAxis': display_schema['metrics'][metric_key].get('axis', 0)
                     if not reference else self._reference_axes_id(reference),
            'color': color
        }

    def xaxis_options(self, dataframe, dim_ordinal, display_schema):
        if isinstance(dataframe.index, pd.DatetimeIndex):
            return {'type': 'datetime'}

        result = {'type': 'categorical'}

        categories = self._make_categories(dataframe, dim_ordinal, display_schema)
        if categories is not None:
            result['categories'] = categories

        return result

    def _validate_dimensions(self, dataframe, dimensions):
        if 1 < len(dimensions) and 1 < len(dataframe.columns):
            raise TransformationException('Cannot transform %s chart.  '
                                          'No more than 1 dimension or 2 dimensions '
                                          'with 1 metric are allowed.' % self.chart_type)

    def _prepare_dataframe(self, dataframe, dim_ordinal, dimensions):
        # Replaces invalid values and unstacks the data frame for line charts.

        # Force all fields to be float (Safer for highcharts)
        dataframe = dataframe.replace([np.inf, -np.inf], np.nan)

        # Unstack multi-indices
        if 1 < len(dimensions):
            unstack_levels = list(self._unstack_levels(list(dimensions.items())[1:], dim_ordinal))
            dataframe = dataframe.unstack(level=unstack_levels)

        return dataframe

    @staticmethod
    def _make_categories(dataframe, dim_ordinal, display_schema):
        if not display_schema['dimensions']:
            return None

        category_dimension = list(display_schema['dimensions'].values())[0]

        if 'display_field' in category_dimension:
            display_field = category_dimension['display_field']
            return dataframe.index.get_level_values(display_field).tolist()

        display_options = category_dimension.get('category_dimension', {})
        return [display_options.get(value, value) for value in dataframe.index]


class HighchartsStackedColumnTransformer(HighchartsColumnTransformer):
    """
    Transformer for a Highcharts Stacked Column chart
    http://www.highcharts.com/demo/column-stacked
    """

    def _make_series_item(self, idx, item, dim_ordinal, display_schema, metrics, reference, color='#000'):
        metric_key = utils.slice_first(idx)
        return {
            'name': self._format_label(idx, dim_ordinal, display_schema, reference),
            'data': self._format_data(item),
            'tooltip': self._format_tooltip(display_schema['metrics'][metric_key]),
            'color': color
        }

    def transform(self, dataframe, display_schema):
        result = super(HighchartsStackedColumnTransformer, self).transform(dataframe, display_schema)
        result['plotOptions'] = result.get('plotOptions', {})
        result['plotOptions'].update({
            'column': {
                'stacking': 'normal'
            }
        })

        return result

    def yaxis_options(self, dataframe, dim_ordinal, display_schema):
        return [{'title': None}]


class HighchartsBarTransformer(HighchartsColumnTransformer):
    """
    Transformer for a Highcharts Bar chart
    http://www.highcharts.com/demo/bar-basic
    """
    chart_type = 'bar'


class HighchartsStackedBarTransformer(HighchartsBarTransformer):
    """
    Transformer for a Highcharts Stacked Bar chart
    http://www.highcharts.com/demo/bar-stacked
    """

    def _make_series_item(self, idx, item, dim_ordinal, display_schema, metrics, reference, color='#000'):
        metric_key = utils.slice_first(idx)
        return {
            'name': self._format_label(idx, dim_ordinal, display_schema, reference),
            'data': self._format_data(item),
            'tooltip': self._format_tooltip(display_schema['metrics'][metric_key]),
            'color': color
        }

    def transform(self, dataframe, display_schema):
        result = super(HighchartsStackedBarTransformer, self).transform(dataframe, display_schema)
        result['plotOptions'] = result.get('plotOptions', {})
        result['plotOptions'].update({
            'series': {
                'stacking': 'normal'
            }
        })

        return result

    def yaxis_options(self, dataframe, dim_ordinal, display_schema):
        return [{'title': None}]


class HighchartsPieTransformer(HighchartsLineTransformer):
    """
    Transformer to create the correct JSON payload for Highcharts Pie Charts
    http://www.highcharts.com/demo/pie-basic
    """
    chart_type = 'pie'

    def prevalidate_request(self, slicer, metrics, dimensions,
                            metric_filters, dimension_filters,
                            references, operations):
        """
        Ensure no references or operations are passed and that there is no more than one enabled metric
        """

        if len(references) > 0 or len(operations) > 0:
            raise TransformationException('References and Operations cannot be used with '
                                          '{} charts'.format(self.chart_type))

        if len(utils.flatten(metrics)) > 1:
            raise TransformationException('Only one metric can be specified when using '
                                          '{} charts'.format(self.chart_type))

    def transform(self, dataframe, display_schema):
        """
        Transform the dataframe into the format Highcharts expects for Pie charts

        :param dataframe: Dataframe containing queried data
        :param display_schema: Dictionary defining how the metrics and dimensions should be displayed
        :return: Dictionary in the required Highcharts Pie chart format ready to be dumped into JSON
        """
        dim_ordinal = {name: ordinal
                       for ordinal, name in enumerate(dataframe.index.names)}
        dataframe = self._prepare_dataframe(dataframe, dim_ordinal, display_schema['dimensions'])
        series = self._make_series(dataframe, dim_ordinal, display_schema)

        # Issues have been seen when showing over 900 data points on a Highcharts pie chart.
        max_data_points = 900
        num_data_points = len(series['data'])
        if num_data_points > max_data_points:
            raise TransformationException('You have reached the maximum number of data points that can be shown '
                                          'on a pie chart. Maximum number of data points: {}. '
                                          'Current number of data points: {}.'.format(max_data_points, num_data_points))

        metric_key = self._get_metric_key(dataframe)

        result = {
            'chart': {'type': self.chart_type},
            'title': {'text': None},
            'tooltip': self._format_tooltip(display_schema['metrics'][metric_key]),
            'series': [series],
            'plotOptions': {
                'pie': {
                    'allowPointSelect': True,  # Allows users to select a piece of pie and it separates out
                    'cursor': 'pointer',
                    'dataLabels': {
                        'enabled': True,
                        'format': '<b>{point.name}</b>: {point.percentage:.1f} %',
                        'style': {
                            'color': COLORS.get(settings.highcharts_colors, 'grid')
                        }
                    }
                }
            },
        }

        return result

    def _prepare_dataframe(self, dataframe, dim_ordinal, dimensions):
        """
        Force all fields to be float (Safer for highcharts)

        :param dataframe: Dataframe containing queried data
        :param dim_ordinal: Dictionary containing dimensions with their associated order
        :param dimensions: OrderedDict of dimensions along with their display options/display field
        :return: Dataframe with float values and without np.inf values
        """
        return dataframe.astype(np.float).replace([np.inf, -np.inf], np.nan)

    @staticmethod
    def _get_metric_key(dataframe):
        """
        Get the metric label (There will only ever be one metric)

        :param dataframe: Dataframe containing queried data
        """
        metrics = list(dataframe.columns.levels[0]
                       if isinstance(dataframe.columns, pd.MultiIndex)
                       else dataframe.columns)
        return metrics[0]

    def _make_series(self, dataframe, dim_ordinal, display_schema, reference=None):
        """
        Create the series. Pie charts only ever have a single series.

        :param dataframe: Dataframe containing queried data
        :param dim_ordinal: Dictionary containing dimensions with their associated order
        :param display_schema: Dictionary defining how the metrics and dimensions should be displayed
        :param reference: Not used - set here to match parent classes' function signature
        :return: Dictonary containing the series name and data list
        """
        metric_key = self._get_metric_key(dataframe)
        return {
            'name': display_schema['metrics'][metric_key].get('label', metric_key),
            'data': [(self._format_label(idx, dim_ordinal, display_schema, reference), _format_data_point(item))
                     for idx, item in dataframe[metric_key].iteritems()]
        }

    def _format_label(self, idx, dim_ordinal, display_schema, reference):
        """
        Create the labels for the pie pieces. If there is more than one dimension,
        these will be shown in the format (dim1, dim2). Otherwise, it will show just the
        individual dimension without parentheses.

        :param idx: Dataframe index
        :param dim_ordinal: Dictionary containing dimensions with their associated order
        :param display_schema: Dictionary defining how the metrics and dimensions should be displayed
        :param reference: Not used - set here to match parent classes' function signature
        :return: Dimension label string
        """
        dimension_labels = [self._format_dimension_display(dim_ordinal, key, dimension, idx)
                            for key, dimension in list(display_schema['dimensions'].items())]

        label = (
            '{dimensions}'.format(
                dimensions=', '.join(map(str, dimension_labels))
            )
            if dimension_labels else ''
        )
        return '({})'.format(label) if len(dimension_labels) > 1 else label



# coding: utf8


class Operation(object):
    """
    The `Operation` class represents an operation in the `Slicer` API.
    """
    key = None
    label = None

    def schemas(self):
        pass

    def metrics(self):
        return []


class Totals(Operation):
    """
    `Operation` for rolling up totals across dimensions in queries.  This will append the totals across a dimension to
    a dimension.
    """
    key = '_total'
    label = 'Total'

    def __init__(self, *dimension_keys):
        self.dimension_keys = dimension_keys


class L1Loss(Operation):
    """
    Performs L1 Loss (mean abs. error) operation on a metric using another metric as the target.
    """
    key = 'l1loss'
    label = 'L1 loss'

    def __init__(self, metric_key, target_metric_key):
        self.metric_key = metric_key
        self.target_metric_key = target_metric_key

    def schemas(self):
        return {
            'key': self.key,
            'metric': self.metric_key,
            'target': self.target_metric_key,
        }

    def metrics(self):
        return [self.metric_key, self.target_metric_key]


class L2Loss(L1Loss):
    """
    Performs L2 Loss (mean sqr. error) operation on a metric using another metric as the target.
    """
    key = 'l2loss'
    label = 'L2 loss'


class CumSum(Operation):
    """
    Accumulates the sum of one or more metrics.
    """
    key = 'cumsum'
    label = 'cum. sum'

    def __init__(self, metric_key):
        self.metric_key = metric_key

    def schemas(self):
        return {
            'key': self.key,
            'metric': self.metric_key,
        }

    def metrics(self):
        return (self.metric_key,)


class CumMean(CumSum):
    """
    Accumulates the mean of one or more metrics
    """
    key = 'cummean'
    label = 'cum. mean'



# coding: utf-8
from .filters import EqualityFilter, ContainsFilter, ExcludesFilter, RangeFilter, WildcardFilter
from .managers import SlicerException
from .schemas import (Slicer, Metric, Dimension, CategoricalDimension, ContinuousDimension, NumericInterval,
                      UniqueDimension, DatetimeDimension, DatetimeInterval, DimensionValue, EqualityOperator, Join)



# coding: utf-8
import pandas as pd


def get_cum_metric(dataframe, schema, reference=None):
    metric = schema['metric']
    if reference is None:
        return dataframe[metric]
    return dataframe[reference, metric]


def get_loss_metric(dataframe, schema, reference=None):
    target, metric = (
        (schema['target'], schema['metric'])
        if reference is None
        else ((reference, schema['target']), (reference, schema['metric']))
    )

    error_series = dataframe[target] - dataframe[metric]
    error_series.name = (schema['metric']
                         if reference is None
                         else (reference, schema['metric']))
    return error_series


FUNCTIONS = {
    'cumsum': (get_cum_metric, lambda x: x.expanding(min_periods=1).sum()),
    'cummean': (get_cum_metric, lambda x: x.expanding(min_periods=1).mean()),
    'l1loss': (get_loss_metric, lambda x: x.abs().expanding(min_periods=1).mean()),
    'l2loss': (get_loss_metric, lambda x: x.pow(2).expanding(min_periods=1).mean()),
}


class OperationManager(object):
    def post_process(self, dataframe, operation_schema):
        dataframe = dataframe.copy()

        for schema in operation_schema:
            key = schema['key']

            value_func, operation_func = FUNCTIONS.get(schema['key'])
            if not value_func or not operation_func:
                continue

            self._perform_operation(dataframe, key, schema, value_func, operation_func)

        return dataframe

    def _perform_operation(self, dataframe, key, schema, value_func, operation):
        # Check for references
        references = (dataframe.columns.get_level_values(0).tolist()
                      if isinstance(dataframe.columns, pd.MultiIndex)
                      else [None])

        for reference in references:
            metric_df = value_func(dataframe, schema, reference=reference)

            operation_key = ('{}_{}'.format(metric_df.name, key)
                             if reference is None
                             else (reference, '{}_{}'.format(metric_df.name[1], key)))

            if isinstance(dataframe.index, pd.MultiIndex):
                unstack_levels = list(range(1, len(dataframe.index.levels)))
                dataframe[operation_key] = metric_df.groupby(level=unstack_levels).apply(operation)

            else:
                dataframe[operation_key] = operation(metric_df)
